/*
  hsreg.h

  PCIE40 registers


*/

#ifndef __PICE40REG_H__
#define __PICE40REG_H__

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <iostream>

#if defined(__cplusplus)
extern "C" {
#endif
#if defined(__dummy_close_bracket_to_cheat_emacs_auto_indent)
}
#endif

#define SLC_BAR 2

#define SLC_BASE_BIT 0
#define SLC_B2LINK_RX_READY1   0x000500E0  // ready state for rxdata, ch 0-23, 1 bit pre channel
#define SLC_B2LINK_RX_READY2   0x00050100  // ready state for rxdata, ch 24-47, 1 bit pre channel
#define SLC_B2LINK_TX_READY1   0x00050120  // ready state for txdata, ch 0-23, 1 bit pre channel
#define SLC_B2LINK_TX_READY2   0x00050140  // ready state for txdata, ch 24-47, 1 bit pre channel
#define SLC_B2LINK_RX_STATUS1  0x00050160  // status of rxlink, ch 0-23, 1 bit pre channel
#define SLC_B2LINK_RX_STATUS2  0x00050180  // status of rxlink, ch 24-47, 1 bit pre channel
#define SLC_B2LINK_TX_STATUS1  0x000501A0  // status of txlink, ch 0-23, 1 bit pre channel
#define SLC_B2LINK_TX_STATUS2  0x000501C0  // status of txlink, ch 24-47, 1 bit pre channel
#define SLC_B2LINK_VETOIDLE    0x00070000  // count vetoidle
#define SLC_B2LINK_VETODATA    0x00070010  // count vetodata
#define SLC_B2LINK_BUFDOWN0    0x00080000  // buffer recent two words of rxdata
#define SLC_B2LINK_BUFDOWN1    0x00090000  // buffer recent third words of rxdata [15-0]+3 charisk[21-16]
#define SLC_B2LINK_CNTDOWN     0x000A0000  //  
#define SLC_B2LINK_MONSTATUS   0x000A0010  // sta_linkdown, vetodata, vetoidle

#define SLC_B2LINK_TIE_STATUS1 0x000501E0  // tied link up/down channel 0-23, 1 bit pre channel
#define SLC_B2LINK_TIE_STATUS2 0x00050200  // tied link up/down channel 24-47, 1 bit pre channel
#define SLC_B2LINK_STATUS1     0x000502E0  // link up/down channel 0-23, 1 bit pre channel
#define SLC_B2LINK_STATUS2     0x00050300  // link up/down channel 24-47, 1 bit pre channel
#define SLC_B2LINK_LANEUP1     0x00050320  // lane up/down channel 0-23, 1 bit pre channel
#define SLC_B2LINK_LANEUP2     0x00050340  // lane up/down channel 24-47, 1 bit pre channel
#define SLC_B2LINK_MASK1       0x00050520  // mask info channel 0-23, 1 bit pre channel
#define SLC_B2LINK_MASK2       0x00050540  // make info channel 24-47, 1 bit pre channel
#define SLC_RFIFO_STATUS       0x00060010

#define SLC_B2LINK_KBYTE       0x00A00000  // count 32bit data pocket: counter bit 47-16
#define SLC_B2LINK_NWORD       0x000F0000  // count 16bit data pocket: counter bit 15-0
#define SLC_B2LINK_DCHECK      0x00A10000  // flags of data check: 0-9bit:NG flag, 16-25:OK flag
#define SLC_B2LINK_EVTSZ       0x000E0010  // Event size
#define SLC_B2LINK_MAXEVTSZ    0x000E0000  // Maximum event size
//#define SLC_B2LINK_NEVT        0x000D0010  // 16 bit fifo_eve_cnt
#define SLC_B2LINK_NEVT        0x00A20000  // counter of nevent
#define SLC_B2LINK_FFUSAGE     0x000D0000  // FIFO usage 
#define SLC_B2LINK_LFFUSAGE    0x000C0010  // Length FIFO usage 
#define SLC_B2LINK_RSTGBT      0x000C0000  // counter of autoreset of gbt 
#define SLC_B2LINK_CRC32       0x000B0000  // counter of crc32 
#define SLC_B2LINK_ALMOSTFULL1 0x00050360  // flag of almost full 0-23 ch
#define SLC_B2LINK_ALMOSTFULL2 0x00050380  // flag of almost full 24-47 ch

#define SLC_PCIE40_WORD2DMA1   0x000503A0  // counter 0-23 bit 
#define SLC_PCIE40_WORD2DMA2   0x000503C0  // counter 24-47 bit
#define SLC_PCIE40_DMA         0x000503E0 
#define SLC_PCIE40_STATUS      0x00050000 
#define SLC_RUM_NUMBER         0x00050020
#define SLC_TRGTAG_NUMBER      0x00050040


int pcie40_b2l_gbt(int dev, int ch);
int pcie40_b2l_status(int dev, int ch);
int pcie40_b2l_tiestatus(int dev, int ch);
int pcie40_b2l_txstatus(int dev, int ch);
int pcie40_b2l_rxstatus(int dev, int ch);
int pcie40_b2l_txready(int dev, int ch);
int pcie40_b2l_rxready(int dev, int ch);
int pcie40_b2l_mask(int dev, int ch);
int pcie40_b2l_set_mask(int dev, int ch, bool mask);
int pcie40_b2l_set_bankmask(int dev, int bank, bool mask);
int pcie40_b2l_cnt_vetoidle(int dev, int ch);
int pcie40_b2l_cnt_vetodata(int dev, int ch);
int pcie40_b2l_3rdrxdata(int dev, int ch);
int pcie40_b2l_2ndrxdata(int dev, int ch);
int pcie40_b2l_1strxdata(int dev, int ch);
int pcie40_b2l_3rdcharisk(int dev, int ch);
int pcie40_b2l_2ndcharisk(int dev, int ch);
int pcie40_b2l_1stcharisk(int dev, int ch);
int pcie40_b2l_cnt_linkdown(int dev, int ch);
int pcie40_b2l_monstatus(int dev, int ch);
int pcie40_b2l_usebadidle(int dev, int ch);
int pcie40_b2l_usebaddata(int dev, int ch);
int pcie40_b2l_bufdata0(int dev, int ch);
int pcie40_b2l_bufdata1(int dev, int ch);
int pcie40_b2l_sta_linkdown(int dev, int ch);
int pcie40_b2l_cntcrc32(int dev, int ch);
int pcie40_b2l_checkFlagOK(int dev, int ch);
int pcie40_b2l_checkFlagNG(int dev, int ch);
int pcie40_b2l_rstgbt(int dev, int ch);
int pcie40_b2l_lffusage(int dev, int ch);
int pcie40_b2l_ffusage(int dev, int ch);
int pcie40_b2l_nevent(int dev, int ch);
int pcie40_b2l_maxevtsz(int dev, int ch);
int pcie40_b2l_evtsz(int dev, int ch);
int pcie40_b2l_fffull(int dev, int ch);
double pcie40_b2l_nkyte(int dev, int ch);
double pcie40_nkbyte_dma(int dev);
int pcie40_status(int dev);
int pcie40_dma_status(int dev);
int pcie40_run_number(int dev);
int pcie40_trgtag_number(int dev);
int pcie40_trg_type(int dev);
int pcie40_memo_status(int dev);
int pcie40_ttd_status(int dev);
int pcie40_ttdclk_status(int dev);
int pcie40_trg_number(int dev);
int pcie40_rxlink_status(int dev);
int pcie40_txlink_status(int dev);
int pcie40_fee_synced(int dev);
int pcie40_disp_status(int dev);
int pcie40_rxdata_error(int dev);



/* FEE registers as implemented in belle2link_cdc */

#define PCIE40REG_SERIAL    0x00
#define PCIE40REG_PDSTL     0x01 /* 0x01 .. 0x60 for 48 ch (x2) */
#define PCIE40REG_PDSTLMIN  0x01
#define PCIE40REG_PDSTLMAX  0x60
#define PCIE40REG_ADCTH     0x61 /* same as lower 8 bit of ADC threshold setting */
#define PCIE40REG_ADCL      0x61
#define PCIE40REG_ADCH      0x62
#define PCIE40REG_WINDOW    0x63
#define PCIE40REG_DELAY     0x64
#define PCIE40REG_TDCTH     0x65 /* same as lower 8 bit of DAC setting for TDC */
#define PCIE40REG_DACL      0x65
#define PCIE40REG_DACH      0x66
#define PCIE40REG_CDCCONT   0x67 /* part of b2ldo (FEE_CONTROL) moved from 0x69 */

/* FEE registers for Belle2link internal use */

#define PCIE40REG_CRCERR    0x68
#define PCIE40REG_FEECONT   0x69
#define PCIE40REG_D32A      0x6a
#define PCIE40REG_D32B      0x6b
#define PCIE40REG_D32C      0x6c
#define PCIE40REG_D32D      0x6d

/* HSLB registers in addition to predefined FINESSE registers */

#define PCIE40REG_D32       0x6e
#define PCIE40REG_A32       0x6f
#define PCIE40REG_STAT      0x72   /* b2ldo (STATUS) */

#define PCIE40REG_CCLK      0x74   /* cpld */
#define PCIE40REG_CONF      0x75   /* cpld */
#define PCIE40REG_CPLDVER   0x76   /* cpld */
#define PCIE40REG_FEEHWTYPE 0x77
#define PCIE40REG_FEESERIAL 0x78
#define PCIE40REG_FEEFWTYPE 0x79
#define PCIE40REG_FEEFWVER  0x7a

#define B2LFEE_UNDEF   0
#define B2LFEE_SVD     1
#define B2LFEE_CDC     2
#define B2LFEE_BPID    3
#define B2LFEE_EPID    4
#define B2LFEE_ECL     5
#define B2LFEE_KLM     6
#define B2LFEE_TRG     7
#define B2LFEE_TEST    15

static const int hsreg_revision = 1;

struct pcie40reg {
  uint32_t stat;    /* 72 */
  uint8_t  csr;     /* 73 */
  uint8_t  xbusy;   /* 74 */
  uint8_t  conf;    /* 75 */
  uint8_t  feehw;   /* 77 >> 4 */
  uint8_t  feefw;   /* 79 >> 4 */
  uint8_t  feecrce; /* 68 */
  uint16_t feeser;  /* 78 | ((77 & f) << 8) */
  uint16_t feever;  /* 7a | ((79 & f) << 8) */
  uint32_t rxdata;  
  uint32_t uptime; 
 
  uint32_t dmastatus; 
  uint16_t eventsz; 
  uint32_t nevent; 
  double   nkbyte;
  double   nkbyte2dma; 
  uint32_t nword;   
  uint32_t vetoset;
  uint32_t vetocnt; 
  uint32_t vetobuf[2];
  uint16_t maxevtsz; 
  uint16_t ffusage; 
  uint16_t lffusage; 
  uint16_t rstgbt; 
  uint32_t cntcrc32; 
  uint16_t cflagOK; 
  uint16_t cflagNG;
  uint16_t vetodata;
  uint16_t vetoidle;
  uint16_t monstatus;
  uint16_t status;
  bool     b2lsta[2];
  bool     b2lgbt[2];
  bool     b2ltxsta[2];
  bool     b2lrxsta[2];
  bool     b2ltxrdy[2];
  bool     b2lrxrdy[2];
  bool     b2lmask[2];
  bool     b2lfful[2];
};

typedef struct pcie40reg pcie40reg_t;


int pcie40reg_read(int fd /* 0-3 */, pcie40reg_t*);
int pcie40reg_getfee(int fd /* 0-3 */, pcie40reg_t*);

/* ---------------------------------------------------------------------- *\
 *    feename
\* ---------------------------------------------------------------------- */
const char *
feename(int hwtype, int fwtype)
{
  static const char *feetype[] = {
    "UNDEF", "SVD", "CDC", "TOP", "ARI", "ECL", "EECL", "KLM", "EKLM",
    "TRG", "FEE10", "FEE11", "FEE12", "FEE13", "DEMO", "TEST" };
  static const char *demotype[] = {
    "UNDEF", "HSLB-B2L", "SP605-B2L", "ML605-B2L", "AC701-B2L" };
  static const char *trgtype[] = {
    "TRGMERGER", /* = 0 */
    "TRGTSF",    /* = 1 */
    "TRG2D",     /* = 2 */
    "TRG3D",     /* = 3 */
    "TRGNN",     /* = 4 */
    "TRGEVTT",   /* = 5 */
    "TRGGRL",    /* = 6 */
    "TRGGDL",    /* = 7 */
    "TRGETM",    /* = 8 */
    "TRGTOP",    /* = 9 */
    "TRGKLM",    /* = 10 */
  };

  hwtype &= 0x0f;
  fwtype  %= 16;

  if (hwtype == 14 && fwtype > 0 && fwtype <= 4) {
    return demotype[fwtype];
  } else if (hwtype == 9 && fwtype >= 0 && fwtype <= 10) {
    return trgtype[fwtype];
  } else {
    int i;
    static int ringi = 0;
    static int ringval[32];
    static char ringbuf[32][20];
    for (i=0; i<32; i++) {
      if (ringval[i] == 0x100 + (hwtype<<4) + fwtype) break;
    }
    if (i == 32) {
      i = ringi;
      ringval[i] = 0x100 + (hwtype<<4) + fwtype;
      sprintf(ringbuf[i], "%s", feetype[hwtype]);
      ringi = (ringi + 1) % 32;
    }
    return ringbuf[i];
  }
}

#if defined(__dummy_open_bracket_to_cheat_emacs_auto_indent)
__dummy_open_bracket_to_cheat_emacs_auto_indent {
#endif
#if defined(__cplusplus)
}
#endif

#endif /* __PCIE40REG_H__ */

